//
// turing - simulation of Turing machines
// Copyright (C) 2009,2010  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "turing.h"

#include <iostream>
#include <sstream>

void check_tm(
    const TM::Turing_machine& tm,
    const TM::Character* inputs[],
    const TM::Character* outputs[],
    const size_t steps[],
    bool debug = false)
{
    size_t index = 0;
    while( inputs[index]) {

        TM::Turing_machine_instance i( tm, inputs[index]);
        if( debug)
            std::cout << "input: " << i.get_tape() << std::endl;
        i.run( debug ? 0 : steps[index]+1, false);

        if( debug) {
            std::cout << "output: " << i.get_tape() << std::endl;
            std::cout << "exp. output: " << outputs[index] << std::endl;
            std::cout << "steps: " << i.get_steps() << std::endl;
        } else {
            std::ostringstream str;
            str << i.get_tape();
            TM_ALWAYS_CHECK(
                (outputs[index] == 0) || (str.str() == outputs[index]));
            TM_ALWAYS_CHECK(
                (steps[index] == 0) || (i.get_steps() == steps[index]));
        }

        ++index;
    }
}

void check_goedel(
    const TM::Turing_machine& tm,
    const char* goedel,
    TM::Encoding encoding,
    bool debug = false)
{
    bool incomplete
        = (tm.get_initial_state() != 0) || (tm.get_halting_state() != 1);

    // check Goedel number (if provided)
    if( debug)
        std::cout << "goedel: " << TM::goedel( tm, encoding, incomplete)
             << std::endl << std::endl;
    else if( goedel) {
        TM_ALWAYS_CHECK( TM::goedel( tm, encoding, incomplete) == goedel);
    }

    // check TM reconstruction from Goedel number (if provided)
    if( !debug && goedel) {
        TM::Turing_machine* tm2
            = TM::ungoedel( goedel, encoding, tm.get_alphabet());
        if( debug)
            if( tm2)
                std::cout << "ungoedel: ok" << std::endl;
            else
                std::cout << "ungoedel: failed" << std::endl;
        else {
            TM_ALWAYS_CHECK( tm2);
            tm2->set_initial_state( tm.get_initial_state());
            tm2->set_halting_state( tm.get_halting_state());
            TM_ALWAYS_CHECK( tm == *tm2);
        }
        delete tm2;
    }
}

/// This TM adds 1 to the input (interpreted as binary number).
void test_add_1()
{
    const TM::Character* inputs[]
        = {  "", "0",  "1", "10",  "11", "00", "01", "010", "011", 0 };
    const TM::Character* outputs[]
        = { "1", "1", "10", "11", "100", "01", "10", "011", "100", 0 };
    const size_t         steps[]
        = {   2,   3,    4,    4,     6,    4,    5,     5,     7, 0 };
    const char*          goedel_f01
        = "1110101010100011"
             "010010100100011"
             "01000100010001011"
             "0001010010010011"
             "0001001000101011"
             "0001000100100100111";
    const char*          goedel_f2
        = "1110101011"
             "01001011"
             "0001000100011"
             "011"
             "0010010011"
             "010100011"
             "00100100111";

    TM::Alphabet a;
    TM::Turing_machine tm( 0, 1, a);
    tm.add_transition( TM::Transition( 0, '0', 0, '0', 'R'));
    tm.add_transition( TM::Transition( 0, '1', 0, '1', 'R'));
    tm.add_transition( TM::Transition( 0, 'B', 2, 'B', 'L'));
    tm.add_transition( TM::Transition( 2, '0', 1, '1', 'N'));
    tm.add_transition( TM::Transition( 2, '1', 2, '0', 'L'));
    tm.add_transition( TM::Transition( 2, 'B', 1, '1', 'N'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f01, TM::ENCODING_F0);
    check_goedel( tm, goedel_f01, TM::ENCODING_F1);
    check_goedel( tm, goedel_f2 , TM::ENCODING_F2);
}

/// This TM subtracts 1 from the input (interpreted as binary number).
void test_sub_1()
{
    const TM::Character* inputs[]
        = { "1", "10",  "11", "100", "101", "110", "111", 0 };
    const TM::Character* outputs[]
        = { "0", "01",  "10", "011", "100", "101", "110", 0 };
    const size_t         steps[]
        = {   3,   5,    4,    7,     5,    6,    5,     0 };
    const char*          goedel_f01
        = "1110101010100011"
             "010010100100011"
             "01000100010001011"
             "0001010001001011"
             "0001001001010011"
             "0001000100100100111";
    const char*          goedel_f2
        = "1110101011"
             "01001011"
             "0001000100011"
             "011"
             "0100100011"
             "001010011"
             "00100100111";

    TM::Alphabet a;
    TM::Turing_machine tm( 0, 1, a);
    tm.add_transition( TM::Transition( 0, '0', 0, '0', 'R'));
    tm.add_transition( TM::Transition( 0, '1', 0, '1', 'R'));
    tm.add_transition( TM::Transition( 0, 'B', 2, 'B', 'L'));
    tm.add_transition( TM::Transition( 2, '0', 2, '1', 'L'));
    tm.add_transition( TM::Transition( 2, '1', 1, '0', 'N'));
    tm.add_transition( TM::Transition( 2, 'B', 1, '1', 'N'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f01, TM::ENCODING_F0);
    check_goedel( tm, goedel_f01, TM::ENCODING_F1);
    check_goedel( tm, goedel_f2 , TM::ENCODING_F2);
}

/// This TM shifts the input by one cell to the left.
void test_leftshift_1()
{
    const TM::Character* inputs[]
        = {  "", "0",  "1", "10",  "11", "00", "01", "010", "011", 0 };
    const TM::Character* outputs[]
        = {  "", "0",  "1", "10",  "11", "00", "01", "010", "011", 0 };
    const size_t         steps[]
        = {   2,   6,    6,   10,    10,   10,   10,    14,    14, 0 };
    const char*          goedel_f0
        = "1110101010100011"
             "010010100100011"
             "01000100010001011"
             "00010100001000100011"
             "0001001000001000100011"
             "0001000100100010011"
             "000010001000000101011"
             "00000100010000001001011"
             "00000010001000100010111";
    const char*          goedel_f1
        = "1110101010100011"
             "010010100100011"
             "01000100010001011"
             "00010101000100011"
             "0001001000001000100011"
             "0001000100100010011"
             "000010001000000101011"
             "000001000101001011"
             "00000010001000100010111";
    const char*          goedel_f2
        = "1110101011"
             "01001011"
             "0001000100011"
             "011"
             "0000011"
             "0000010001011"
             "00100010011"
             "0011"
             "0011"
             "00000010100011"
             "0011"
             "0011"
             "000000100100011"
             "0011"
             "0011"
             "00010001000111";

    TM::Alphabet a;
    TM::Turing_machine tm( 0, 1, a);
    tm.add_transition( TM::Transition( 0, '0', 0, '0', 'R'));
    tm.add_transition( TM::Transition( 0, '1', 0, '1', 'R'));
    tm.add_transition( TM::Transition( 0, 'B', 2, 'B', 'L'));
    tm.add_transition( TM::Transition( 2, '0', 3, 'B', 'R'));
    tm.add_transition( TM::Transition( 2, '1', 4, 'B', 'R'));
    tm.add_transition( TM::Transition( 2, 'B', 1, 'B', 'N'));
    tm.add_transition( TM::Transition( 3, 'B', 5, '0', 'L'));
    tm.add_transition( TM::Transition( 4, 'B', 5, '1', 'L'));
    tm.add_transition( TM::Transition( 5, 'B', 2, 'B', 'L'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f0, TM::ENCODING_F0);
    check_goedel( tm, goedel_f1, TM::ENCODING_F1);
    check_goedel( tm, goedel_f2, TM::ENCODING_F2);
}

/// This TM mirrors the input.
void test_mirror()
{
    const TM::Character* inputs[]
        = {  "", "0",  "1", "10",  "11", "00", "01", "010", "011", 0 };
    const TM::Character* outputs[]
        = {  "", "0",  "1", "01",  "11", "00", "10", "010", "110", 0 };
    const size_t         steps[]
        = {   1,   7,    7,   17,    17,   17,   17,    31,    31, 0 };
    const char*          goedel_f0
        = "1110101000100001011"
             "010010000100001011"
             "01000100100010011"
             "000101000101011"
             "00010010001001011"
             "000100001000100001011"
             "00010001000001010011"
             "00001010000101011"
             "0000100100001001011"
             "00001000010000100001011"
             "0000100010000010010011"
             "000001010000010100011"
             "00000100100000100100011"
             "0000010000100000010000100011"
             "00000010101010011"
             "0000001001010010011"
             "00000010000100000010000100011"
             "00000010001000000010001011"
             "0000000101001010011"
             "000000010010010010011"
             "00000001000010000000100010111";

    TM::Alphabet a( "01B#");
    TM::Turing_machine tm( 0, 1, a);
    tm.add_transition( TM::Transition( 0, '0', 2, '#', 'L'));
    tm.add_transition( TM::Transition( 0, '1', 3, '#', 'L'));
    tm.add_transition( TM::Transition( 0, 'B', 1, 'B', 'N'));
    tm.add_transition( TM::Transition( 2, '0', 2, '0', 'L'));
    tm.add_transition( TM::Transition( 2, '1', 2, '1', 'L'));
    tm.add_transition( TM::Transition( 2, '#', 2, '#', 'L'));
    tm.add_transition( TM::Transition( 2, 'B', 4, '0', 'N'));
    tm.add_transition( TM::Transition( 3, '0', 3, '0', 'L'));
    tm.add_transition( TM::Transition( 3, '1', 3, '1', 'L'));
    tm.add_transition( TM::Transition( 3, '#', 3, '#', 'L'));
    tm.add_transition( TM::Transition( 3, 'B', 4, '1', 'N'));
    tm.add_transition( TM::Transition( 4, '0', 4, '0', 'R'));
    tm.add_transition( TM::Transition( 4, '1', 4, '1', 'R'));
    tm.add_transition( TM::Transition( 4, '#', 5, '#', 'R'));
    tm.add_transition( TM::Transition( 5, '0', 0, '0', 'N'));
    tm.add_transition( TM::Transition( 5, '1', 0, '1', 'N'));
    tm.add_transition( TM::Transition( 5, '#', 5, '#', 'R'));
    tm.add_transition( TM::Transition( 5, 'B', 6, 'B', 'L'));
    tm.add_transition( TM::Transition( 6, '0', 1, '0', 'N'));
    tm.add_transition( TM::Transition( 6, '1', 1, '1', 'N'));
    tm.add_transition( TM::Transition( 6, '#', 6, 'B', 'L'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f0, TM::ENCODING_F0);
    // The F1 and F2 encodings cannot handle transitions
    // to state 0 except from state 0.
}

/// This TM is a 3-state, 2-symbol busy beaver.
void test_busy_beaver_1()
{
    const TM::Character* inputs[]  = { "", 0 };
    const TM::Character* outputs[] = { "111111", 0 };
    const size_t         steps[]   = { 14, 0 };
    const char*          goedel_f0 = "1110100100010100011"
                                        "01010010100011"
                                        "00010010000100100011"
                                        "00010100010100011"
                                        "000010010000101011"
                                        "000010101010111";

    TM::Alphabet a( "1B");
    TM::Turing_machine tm( 0, 1, a);
    tm.add_transition( TM::Transition( 0, 'B', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 0, '1', 1, '1', 'R'));
    tm.add_transition( TM::Transition( 2, 'B', 3, 'B', 'R'));
    tm.add_transition( TM::Transition( 2, '1', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 3, 'B', 3, '1', 'L'));
    tm.add_transition( TM::Transition( 3, '1', 0, '1', 'L'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f0, TM::ENCODING_F0);
    // The F1 and F2 encodings cannot handle transitions
    // to state 0 except from state 0.
}

/// This TM is a 4-state, 2-symbol busy beaver.
void test_busy_beaver_2()
{
    const TM::Character* inputs[]  = { "", 0 };
    const TM::Character* outputs[] = { "1B111111111111", 0 };
    const size_t         steps[]   = { 107, 0 };
    const char*          goedel_f0 = "1110100100010100011"
                                        "0101000101011"
                                        "00010010101011"
                                        "000101000001001011"
                                        "00001001000010100011"
                                        "00001010100100011"
                                        "0000010010010100011"
                                        "0000010100001010111";

    TM::Alphabet a( "1B");
    TM::Turing_machine tm( 0, 1, a);
    tm.add_transition( TM::Transition( 0, 'B', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 0, '1', 2, '1', 'L'));
    tm.add_transition( TM::Transition( 2, 'B', 0, '1', 'L'));
    tm.add_transition( TM::Transition( 2, '1', 4, 'B', 'L'));
    tm.add_transition( TM::Transition( 3, 'B', 3, '1', 'R'));
    tm.add_transition( TM::Transition( 3, '1', 0, 'B', 'R'));
    tm.add_transition( TM::Transition( 4, 'B', 1, '1', 'R'));
    tm.add_transition( TM::Transition( 4, '1', 3, '1', 'L'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f0, TM::ENCODING_F0);
    // The F1 and F2 encodings cannot handle transitions
    // to state 0 except from state 0.
}

/// This 5-state, 2-symbol TM is known not to be a busy beaver, because it is
/// known not to produce the largest number of 1's as output.
void test_busy_beaver_3()
{
    const TM::Character* inputs[]  = { "", 0 };
    const TM::Character* outputs[] = { "", 0 };
    const size_t         steps[]   = { 11798826, 0 };
    const char*          goedel_f0 = "11101001000101011"
                                        "01010101011"
                                        "0001001000010100011"
                                        "00010100010100011"
                                        "000010010101011"
                                        "00001010000010100011"
                                        "0000010010101011"
                                        "0000010100000010100011"
                                        "000000100100101011"
                                        "00000010100001001000111";

    size_t output0_len = 1 + 3*2047 + 3;
    TM::Character* output0 = new TM::Character[ output0_len + 1];
    for( size_t index = 0; index < output0_len; ++index)
        output0[index]
            = (index % 3 == 0) && (index >= 3) && (index < output0_len-1)
            ? 'B' : '1';
    output0[output0_len] = '\0';
    outputs[0] = output0;

    TM::Alphabet a( "1B");
    TM::Turing_machine tm( 0, 1, a);
    tm.add_transition( TM::Transition( 0, 'B', 2, '1', 'L'));
    tm.add_transition( TM::Transition( 0, '1', 0, '1', 'L'));
    tm.add_transition( TM::Transition( 2, 'B', 3, '1', 'R'));
    tm.add_transition( TM::Transition( 2, '1', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 3, 'B', 0, '1', 'L'));
    tm.add_transition( TM::Transition( 3, '1', 4, '1', 'R'));
    tm.add_transition( TM::Transition( 4, 'B', 0, '1', 'L'));
    tm.add_transition( TM::Transition( 4, '1', 5, '1', 'R'));
    tm.add_transition( TM::Transition( 5, 'B', 1, '1', 'L'));
    tm.add_transition( TM::Transition( 5, '1', 3, 'B', 'R'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f0, TM::ENCODING_F0);
    // The F1 and F2 encodings cannot handle transitions
    // to state 0 except from state 0.

    outputs[0] = 0;
    delete[] output0;
}

/// This 5-state, 2-symbol TM is possibly a busy beaver. It is currently the
/// TM with the largest number of 1's as output known.
void test_busy_beaver_4()
{
    const TM::Character* inputs[]  = { "", 0 };
    const TM::Character* outputs[] = { "", 0 };
    const size_t         steps[]   = { 47176870, 0 };
    const char*          goedel_f0 = "1110100100010100011"
                                        "01010000101011"
                                        "0001001000010100011"
                                        "00010100010100011"
                                        "000010010000010100011"
                                        "00001010000001001011"
                                        "0000010010101011"
                                        "0000010100000101011"
                                        "00000010010010100011"
                                        "000000101010010111";
                                         
    size_t output0_len = 1 + 3*4095 + 3;
    TM::Character* output0 = new TM::Character[ output0_len + 1];
    for( size_t index = 0; index < output0_len; ++index)
        output0[index]
            = (index % 3 != 2) && (index >= 1) && (index < output0_len-1)
            ? 'B' : '1';
    output0[output0_len] = '\0';
    outputs[0] = output0;

    TM::Alphabet a( "1B");
    TM::Turing_machine tm( 0, 1, a);
    tm.add_transition( TM::Transition( 0, 'B', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 0, '1', 3, '1', 'L'));
    tm.add_transition( TM::Transition( 2, 'B', 3, '1', 'R'));
    tm.add_transition( TM::Transition( 2, '1', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 3, 'B', 4, '1', 'R'));
    tm.add_transition( TM::Transition( 3, '1', 5, 'B', 'L'));
    tm.add_transition( TM::Transition( 4, 'B', 0, '1', 'L'));
    tm.add_transition( TM::Transition( 4, '1', 4, '1', 'L'));
    tm.add_transition( TM::Transition( 5, 'B', 1, '1', 'R'));
    tm.add_transition( TM::Transition( 5, '1', 0, 'B', 'L'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f0, TM::ENCODING_F0);
    // The F1 and F2 encodings cannot handle transitions
    // to state 0 except from state 0.

    outputs[0] = 0;
    delete[] output0;
}

/// This TM is from [MG02], Fig. 1, top left.
void test_margenstern_1()
{
    const TM::Character* inputs[]  = {  "", 0 };
    const TM::Character* outputs[] = {  "0000000000000", 0 };
    const size_t         steps[]   = {   98, 0 };
    const char*          goedel_f0 = "111010001000010100011"
                                        "00100010000000000100010011"
                                        "00101000100100011"
                                        "00100100101011"
                                        "0001000100101011"
                                        "000101000100100011"
                                        "0001001000100100011"
                                        "0000100010000010100011"
                                        "000001000100000010100011"
                                        "00000010001000000010100011"
                                        "0000000100010000000010100011"
                                        "000000001000100000000010100011"
                                        "00000000010001001010111";
    const char*          goedel_f1 = "111010001000010100011"
                                        "00100010000000000100010011"
                                        "001010100100011"
                                        "00100100101011"
                                        "0001000100101011"
                                        "000101000100100011"
                                        "0001001000100100011"
                                        "000010001010100011"
                                        "0000010001010100011"
                                        "00000010001010100011"
                                        "000000010001010100011"
                                        "0000000010001010100011"
                                        "00000000010001001010111";
    const char*          goedel_f2 = "1110011"
                                        "0011"
                                        "0000101011"
                                        "000011"
                                        "010100011"
                                        "0000000000100010011"
                                        "01001011"
                                        "01001011"
                                        "0010100011"
                                        "0011"
                                        "0011"
                                        "00011"
                                        "0011"
                                        "0011"
                                        "00011"
                                        "0011"
                                        "0011"
                                        "00011"
                                        "0011"
                                        "0011"
                                        "00011"
                                        "0011"
                                        "0011"
                                        "00011"
                                        "0011"
                                        "0011"
                                        "00101000111";

    TM::Alphabet a( "01B");
    TM::Turing_machine tm( 0, 9, a);
    tm.add_transition( TM::Transition( 0, 'B', 3, '0', 'R'));
    tm.add_transition( TM::Transition( 1, 'B', 9, 'B', 'N'));
    tm.add_transition( TM::Transition( 1, '0', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 1, '1', 1, '0', 'L'));
    tm.add_transition( TM::Transition( 2, 'B', 1, '0', 'L'));
    tm.add_transition( TM::Transition( 2, '0', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 2, '1', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 3, 'B', 4, '0', 'R'));
    tm.add_transition( TM::Transition( 4, 'B', 5, '0', 'R'));
    tm.add_transition( TM::Transition( 5, 'B', 6, '0', 'R'));
    tm.add_transition( TM::Transition( 6, 'B', 7, '0', 'R'));
    tm.add_transition( TM::Transition( 7, 'B', 8, '0', 'R'));
    tm.add_transition( TM::Transition( 8, 'B', 1, '0', 'L'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f0, TM::ENCODING_F0);
    check_goedel( tm, goedel_f1, TM::ENCODING_F1);
    check_goedel( tm, goedel_f2, TM::ENCODING_F2);
}

/// This TM is from [MG02], Fig. 1, top right.
void test_margenstern_2()
{
    const TM::Character* inputs[]  = {  "", 0 };
    const TM::Character* outputs[] = {  "00000000000000", 0 };
    const size_t         steps[]   = { 104, 0 };
    const char*          goedel_f0 = "1110100010000010100011"
                                        "01010101011"
                                        "00100010001000100011"
                                        "001010010100011"
                                        "000100010000101011"
                                        "00010100010100011"
                                        "000010001010001011"
                                        "00001010000101011"
                                        "000001000100000010100011"
                                        "00000101001000100011"
                                        "00000010001000000010100011"
                                        "00000010100000000001010011"
                                        "0000000100010000000010100011"
                                        "000000001000100000000010100011"
                                        "0000000001000101010111";

    TM::Alphabet a( "01B");
    TM::Turing_machine tm( 0, 9, a);
    tm.add_transition( TM::Transition( 0, 'B', 4, '0', 'R'));
    tm.add_transition( TM::Transition( 0, '0', 0, '0', 'L'));
    tm.add_transition( TM::Transition( 1, 'B', 2, 'B', 'R'));
    tm.add_transition( TM::Transition( 1, '0', 1, '0', 'R'));
    tm.add_transition( TM::Transition( 2, 'B', 3, '0', 'L'));
    tm.add_transition( TM::Transition( 2, '0', 2, '0', 'R'));
    tm.add_transition( TM::Transition( 3, 'B', 0, 'B', 'L'));
    tm.add_transition( TM::Transition( 3, '0', 3, '0', 'L'));
    tm.add_transition( TM::Transition( 4, 'B', 5, '0', 'R'));
    tm.add_transition( TM::Transition( 4, '0', 1, 'B', 'R'));
    tm.add_transition( TM::Transition( 5, 'B', 6, '0', 'R'));
    tm.add_transition( TM::Transition( 5, '0', 9, '0', 'N'));
    tm.add_transition( TM::Transition( 6, 'B', 7, '0', 'R'));
    tm.add_transition( TM::Transition( 7, 'B', 8, '0', 'R'));
    tm.add_transition( TM::Transition( 8, 'B', 0, '0', 'L'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f0, TM::ENCODING_F0);
    // The F1 and F2 encodings cannot handle transitions
    // to state 0 except from state 0.
}

/// This TM is from [MG02], Fig. 1, bottom.
void test_margenstern_3()
{
    const TM::Character* inputs[]  = {  "", 0 };
    const TM::Character* outputs[] = {  "0000000", 0 };
    const size_t         steps[]   = {  32, 0 };
    const char*          goedel_f0 = "111010001000010100011"
                                        "0010001000000100010011"
                                        "00101000100100011"
                                        "00100100101011"
                                        "0001000100101011"
                                        "000101000100100011"
                                        "0000100010000010100011"
                                        "0000010001000101000111";
    const char*          goedel_f1 = "111010001000010100011"
                                        "0010001000000100010011"
                                        "001010100100011"
                                        "00100100101011"
                                        "0001000100101011"
                                        "000101000100100011"
                                        "000010001010100011"
                                        "0000010001000101000111";
    const char*          goedel_f2 = "1110011"
                                        "0011"
                                        "0000101011"
                                        "000011"
                                        "010100011"
                                        "000000100010011"
                                        "01001011"
                                        "0011"
                                        "0010100011"
                                        "0011"
                                        "0011"
                                        "00011"
                                        "0011"
                                        "0011"
                                        "0001010111";

    TM::Alphabet a( "01B");
    TM::Turing_machine tm( 0, 5, a);
    tm.add_transition( TM::Transition( 0, 'B', 3, '0', 'R'));
    tm.add_transition( TM::Transition( 1, 'B', 5, 'B', 'N'));
    tm.add_transition( TM::Transition( 1, '0', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 1, '1', 1, '0', 'L'));
    tm.add_transition( TM::Transition( 2, 'B', 1, '0', 'L'));
    tm.add_transition( TM::Transition( 2, '0', 2, '1', 'R'));
    tm.add_transition( TM::Transition( 3, 'B', 4, '0', 'R'));
    tm.add_transition( TM::Transition( 4, 'B', 2, '0', 'R'));

    check_tm( tm, inputs, outputs, steps);
    check_goedel( tm, goedel_f0, TM::ENCODING_F0);
    check_goedel( tm, goedel_f1, TM::ENCODING_F1);
    check_goedel( tm, goedel_f2, TM::ENCODING_F2);
}

void test_compute_transitions()
{
    const TM::Character* strings[]
        = { "", "0", "1", "00", "01", "10", "11", "000", "111", 0 };
    const size_t         steps[]
        = {  2,   3,   3,    4,    4,    4,    4,     5,     5, 0 };

    TM::State initial = 3;
    TM::State stride = 2;
    size_t index = 0;
    while( strings[index]) {

        TM::Alphabet a;
        TM::Turing_machine tm( 0, 1, a);
        tm.add_transition( TM::Transition( 0, 'B', initial, 'B', 'N'));
        tm.add_transitions_for_fixed_string( initial, stride, strings[index]);
        tm.add_transition( TM::Transition(
            initial + stride * strlen(strings[index]), 'B', 1, 'B', 'N'));

        const TM::Character* inputs[]  = { "", 0 };
        const TM::Character* outputs[] = { "", 0 };
        size_t               steps2[]  = {  0, 0 };
        outputs[0] = strings[index];
        steps2[0] = steps[index];

        check_tm( tm, inputs, outputs, steps2, 0);
        ++index;
    }
}

int main( int /* argc */, char* /* argv */[]) {

    test_add_1();
    test_sub_1();
    test_leftshift_1();
    test_mirror();
    test_busy_beaver_1();
    test_busy_beaver_2();
    test_busy_beaver_3();
    test_busy_beaver_4();
    test_margenstern_1();
    test_margenstern_2();
    test_margenstern_3();

    test_compute_transitions();

    return 0;
}
