//
// turing - simulation of Turing machines
// Copyright (C) 2009,2010  Joachim Reichel <joachim.reichel@gmx.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include "turing_machine_instance.h"

#include "timer.h"

#include <iostream>
#include <cmath>

namespace TM {



Turing_machine_instance::Turing_machine_instance(
    const Turing_machine& turing_machine, const Tape& tape)
  : m_turing_machine( turing_machine),
    m_tape( tape),
    m_state( turing_machine.get_initial_state()),
    m_head( tape.get_left_boundary()),
    m_steps( 0)
{
    TM_ALWAYS_CHECK( m_tape.is_valid( m_turing_machine.get_alphabet()));
    m_turing_machine.optimize();
}

void Turing_machine_instance::step()
{
    if( has_stopped())
        return;

    Character character = m_tape.get_character( m_head);
    size_t index = m_turing_machine.find_transition( m_state, character);

#ifndef NDEBUG
    if( index == static_cast<size_t>( -1)) {
        std::cout << "error (no transition matches):" << *this << std::endl;
        TM_ALWAYS_CHECK( false);
    }
#endif

    const Transition& transition = m_turing_machine.get_transition( index);
    TM_EXPENSIVE_CHECK( transition.matches( m_state, character));

    m_state = transition.get_new_state();
    m_tape.set_character( m_head, transition.get_new_character());

    const Direction& d = transition.get_direction();
    if( d == Left)
        --m_head;
    else if( d == Neutral)
        ;
    else if( d == Right)
        ++m_head;
    else {
        TM_EXPENSIVE_CHECK( false);
    }

    ++m_steps;
}

void Turing_machine_instance::run( size_t steps, bool trace)
{
    bool until_halting_state = (steps == 0);
    do {
        step();
        if( trace)
             std::cout << *this;
    } while( !has_stopped() && (until_halting_state || (--steps != 0)));
}

bool Turing_machine_instance::operator==(
    const Turing_machine_instance& rhs) const
{
    return ( m_turing_machine == rhs.m_turing_machine)
        && ( m_tape == rhs.m_tape)
        && ( m_state == rhs.m_state)
        && ( m_head == rhs.m_head)
        && ( m_steps == rhs.m_steps);
}

bool Turing_machine_instance::operator!=(
    const Turing_machine_instance& rhs) const
{
    return !( *this == rhs);
}

std::ostream& operator<<(
    std::ostream& stream, const Turing_machine_instance& tmi)
{
    stream << tmi.m_steps << " " << tmi.m_state << " ";
    ptrdiff_t left_boundary  = tmi.m_tape.get_left_boundary();
    ptrdiff_t right_boundary = tmi.m_tape.get_right_boundary();
    left_boundary  = std::min( left_boundary,  tmi.m_head);
    right_boundary = std::max( right_boundary, tmi.m_head);

    for( ptrdiff_t index = left_boundary; index < tmi.m_head; ++index)
        stream << tmi.m_tape.get_character( index);
    stream << "|" << tmi.m_tape.get_character( tmi.m_head) << "|";
    for( ptrdiff_t index = tmi.m_head+1; index <= right_boundary; ++index)
        stream << tmi.m_tape.get_character( index);
    stream << std::endl;
    return stream;
}



} // namespace TM
